"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var compile_dts_1 = require("./compile-dts");
var types_usage_evaluator_1 = require("./types-usage-evaluator");
var node_modules_helpers_1 = require("./node-modules-helpers");
var logger_1 = require("./logger");
var skippedNodes = [
    ts.SyntaxKind.ExportDeclaration,
    ts.SyntaxKind.ExportAssignment,
    ts.SyntaxKind.ImportDeclaration,
    ts.SyntaxKind.ImportEqualsDeclaration,
];
// tslint:disable-next-line:cyclomatic-complexity
function generateDtsBundle(filePath, options) {
    if (options === void 0) { options = {}; }
    var inlinedLibraries = options.inlinedLibraries || [];
    var importedLibraries = options.importedLibraries || [];
    var allowedTypesLibs = options.allowedTypesLibraries;
    if (!ts.sys.fileExists(filePath)) {
        throw new Error("File \"" + filePath + "\" does not exist");
    }
    var program = compile_dts_1.compileDts(filePath);
    var typeChecker = program.getTypeChecker();
    // we do not need all files from node_modules dir
    var sourceFiles = program.getSourceFiles().filter(function (file) {
        var fileName = file.fileName;
        var libraryName = node_modules_helpers_1.getLibraryName(fileName);
        if (libraryName === null) {
            return true;
        }
        var typesLibName = node_modules_helpers_1.getTypesLibraryName(fileName);
        if (typesLibName !== null && (allowedTypesLibs === undefined || allowedTypesLibs.indexOf(typesLibName) !== -1)) {
            return true;
        }
        return inlinedLibraries.indexOf(libraryName) !== -1 || importedLibraries.indexOf(libraryName) !== -1;
    });
    logger_1.verboseLog("Input source files:\n  " + sourceFiles.map(function (file) { return file.fileName; }).join('\n  '));
    var typesUsageEvaluator = new types_usage_evaluator_1.TypesUsageEvaluator(sourceFiles, typeChecker);
    var rootSourceFileSymbol = typeChecker.getSymbolAtLocation(getRootSourceFile(program));
    if (rootSourceFileSymbol === undefined) {
        throw new Error('Symbol for root source file not found');
    }
    var rootFileExports = typeChecker.getExportsOfModule(rootSourceFileSymbol).map(function (symbol) {
        if (symbol.flags & ts.SymbolFlags.Alias) {
            // so we need to have original symbols from source file
            symbol = typeChecker.getAliasedSymbol(symbol);
        }
        return symbol;
    });
    var usedTypes = new Set();
    var importedSymbols = new Map();
    var resultOutput = '';
    for (var _i = 0, sourceFiles_1 = sourceFiles; _i < sourceFiles_1.length; _i++) {
        var sourceFile = sourceFiles_1[_i];
        logger_1.verboseLog("\n\n======= Preparing file: " + sourceFile.fileName + " =======");
        var sourceFileText = sourceFile.getFullText();
        var fileOutput = '';
        for (var _a = 0, _b = sourceFile.statements; _a < _b.length; _a++) {
            var node = _b[_a];
            // we should skip import and exports statements
            if (skippedNodes.indexOf(node.kind) !== -1) {
                continue;
            }
            var isNodeUsed = false;
            if (types_usage_evaluator_1.isNodeDeclaration(node)) {
                isNodeUsed = rootFileExports.some(typesUsageEvaluator.isTypeUsedBySymbol.bind(typesUsageEvaluator, node));
            }
            else if (node.kind === ts.SyntaxKind.VariableStatement) {
                var declarations = node.declarationList.declarations;
                isNodeUsed = declarations.some(function (declaration) {
                    return isDeclarationExported(rootFileExports, typeChecker, declaration);
                });
            }
            if (!isNodeUsed) {
                logger_1.verboseLog("Skip file member: " + node.getText().replace(/(\n|\r)/g, '').slice(0, 50) + "...");
                continue;
            }
            var typesLibraryName = node_modules_helpers_1.getTypesLibraryName(sourceFile.fileName);
            if (typesLibraryName !== null) {
                if (!usedTypes.has(typesLibraryName)) {
                    logger_1.normalLog("Library \"" + typesLibraryName + "\" will be added via reference directive");
                    usedTypes.add(typesLibraryName);
                }
                break;
            }
            var importedLibraryName = node_modules_helpers_1.getLibraryName(sourceFile.fileName);
            if (importedLibraryName !== null && importedLibraries.indexOf(importedLibraryName) !== -1) {
                var nodeIdentifier = node.name;
                if (nodeIdentifier === undefined) {
                    throw new Error("Import/usage unnamed declaration: " + node.getText());
                }
                var importName = nodeIdentifier.getText();
                logger_1.normalLog("Add import with name \"" + importName + "\" for library \"" + importedLibraryName + "\"");
                var libraryImports = importedSymbols.get(importedLibraryName);
                if (libraryImports === undefined) {
                    libraryImports = new Set();
                    importedSymbols.set(importedLibraryName, libraryImports);
                }
                libraryImports.add(importName);
                continue;
            }
            var nodeText = node.getText();
            var hasNodeExportKeyword = hasNodeModifier(node, ts.SyntaxKind.ExportKeyword);
            var shouldNodeHasExportKeyword = true;
            if (node.kind === ts.SyntaxKind.ClassDeclaration || node.kind === ts.SyntaxKind.EnumDeclaration) {
                if (options.failOnClass === true && node.kind === ts.SyntaxKind.ClassDeclaration) {
                    var classDecl = node;
                    var className = classDecl.name ? classDecl.name.text : '';
                    var errorMessage = "Class was found in generated dts.\n " + className + " from " + sourceFile.fileName;
                    throw new Error(errorMessage);
                }
                // not all classes and enums can be exported - only exported from root file
                shouldNodeHasExportKeyword = isDeclarationExported(rootFileExports, typeChecker, node);
                if (node.kind === ts.SyntaxKind.EnumDeclaration) {
                    // but const enum always can be exported
                    shouldNodeHasExportKeyword = shouldNodeHasExportKeyword || hasNodeModifier(node, ts.SyntaxKind.ConstKeyword);
                }
            }
            nodeText = getTextAccordingExport(nodeText, hasNodeExportKeyword, shouldNodeHasExportKeyword);
            // add jsdoc for exported nodes only
            if (shouldNodeHasExportKeyword) {
                var start = node.getStart();
                var jsDocStart = node.getStart(undefined, true);
                var nodeJSDoc = sourceFileText.substring(jsDocStart, start).trim();
                if (nodeJSDoc.length !== 0) {
                    nodeText = nodeJSDoc + "\n" + nodeText;
                }
            }
            fileOutput += spacesToTabs(nodeText) + "\n";
        }
        if (fileOutput.length === 0) {
            logger_1.verboseLog("No output for file: " + sourceFile.fileName);
        }
        if (options.outputFilenames) {
            fileOutput = "// File: " + sourceFile.fileName + "\n\n" + fileOutput + "\n";
        }
        resultOutput += fileOutput;
    }
    if (importedSymbols.size !== 0) {
        var importsArray = [];
        // we need to have sorted imports of libraries to have more "stable" output
        var sortedEntries = Array.from(importedSymbols.entries()).sort(function (firstEntry, secondEntry) {
            return firstEntry[0].localeCompare(secondEntry[0]);
        });
        for (var _c = 0, sortedEntries_1 = sortedEntries; _c < sortedEntries_1.length; _c++) {
            var entry = sortedEntries_1[_c];
            var libraryName = entry[0], libraryImports = entry[1];
            importsArray.push(generateImport(libraryName, Array.from(libraryImports)));
        }
        resultOutput = importsArray.join('\n') + "\n\n" + resultOutput;
    }
    if (usedTypes.size !== 0) {
        var header = generateReferenceTypesDirective(Array.from(usedTypes));
        resultOutput = header + "\n\n" + resultOutput;
    }
    return resultOutput;
}
exports.generateDtsBundle = generateDtsBundle;
function getRootSourceFile(program) {
    var rootFiles = program.getRootFileNames();
    if (rootFiles.length !== 1) {
        logger_1.verboseLog("Root files:\n  " + rootFiles.join('\n  '));
        throw new Error("There is not one root file - " + rootFiles.length);
    }
    return program.getSourceFile(rootFiles[0]);
}
function isDeclarationExported(exportedSymbols, typeChecker, declaration) {
    if (declaration.name === undefined) {
        return false;
    }
    var declarationSymbol = typeChecker.getSymbolAtLocation(declaration.name);
    return exportedSymbols.some(function (rootExport) { return rootExport === declarationSymbol; });
}
function getTextAccordingExport(nodeText, isNodeExported, shouldNodeBeExported) {
    if (shouldNodeBeExported && !isNodeExported) {
        return 'export ' + nodeText;
    }
    else if (isNodeExported && !shouldNodeBeExported) {
        return nodeText.slice('export '.length);
    }
    return nodeText;
}
function hasNodeModifier(node, modifier) {
    return Boolean(node.modifiers && node.modifiers.some(function (nodeModifier) { return nodeModifier.kind === modifier; }));
}
function spacesToTabs(text) {
    return text.replace(/^(    )+/gm, function (substring) {
        return '\t'.repeat(substring.length / 4);
    });
}
function generateImport(libraryName, imports) {
    // sort to make output more "stable"
    return "import { " + imports.sort().join(', ') + " } from '" + libraryName + "';";
}
function generateReferenceTypesDirective(libraries) {
    return libraries.sort().map(function (library) {
        return "/// <reference types=\"" + library + "\" />";
    }).join('\n');
}
