"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var path = require("path");
var ts = require("typescript");
var argparse_1 = require("argparse");
var bundle_generator_1 = require("./bundle-generator");
var logger_1 = require("./logger");
var check_diagnostics_errors_1 = require("./check-diagnostics-errors");
var get_compiler_options_1 = require("./get-compiler-options");
var parser = new argparse_1.ArgumentParser();
parser.addArgument(['-o', '--out-file'], {
    dest: 'outFile',
    help: 'File name of generated d.ts',
    required: false,
});
parser.addArgument(['-v', '--verbose'], {
    action: 'storeTrue',
    defaultValue: false,
    help: 'Enable verbose logging',
});
parser.addArgument(['--no-check'], {
    action: 'storeTrue',
    defaultValue: false,
    dest: 'noCheck',
    help: 'Skip validation of generated d.ts file',
    type: Boolean,
});
parser.addArgument(['--output-source-file'], {
    action: 'storeTrue',
    defaultValue: false,
    dest: 'outputSourceFileName',
    help: 'Add comment with file path the definitions came from',
    type: Boolean,
});
parser.addArgument(['--fail-on-class'], {
    action: 'storeTrue',
    defaultValue: false,
    dest: 'failOnClass',
    help: 'Fail if generated dts contains class declaration',
    type: Boolean,
});
parser.addArgument(['--external-inlines'], {
    action: 'store',
    dest: 'externalInlines',
    help: 'Comma-separated packages from node_modules to inline typings from it. Used types will be just inlined into output file',
    required: false,
    type: String,
});
parser.addArgument(['--external-imports'], {
    action: 'store',
    dest: 'externalImports',
    help: 'Comma-separated packages from node_modules to import typings from it. Used types will be imported by "import { First, Second } from \'library-name\';"',
    required: false,
    type: String,
});
parser.addArgument(['--external-types'], {
    action: 'store',
    dest: 'externalTypes',
    help: 'Comma-separated packages from @types to import typings from it via triple-slash reference directive. By default all packages are allowed and will be used according their usages',
    required: false,
    type: String,
});
parser.addArgument(['--config'], {
    action: 'store',
    dest: 'config',
    help: 'File path to generator config file',
    required: false,
    type: String,
});
parser.addArgument(['file'], { nargs: 1 });
// tslint:disable-next-line:no-any
function configToArgs(config, inFile) {
    var result = [];
    for (var _i = 0, _a = Object.keys(config); _i < _a.length; _i++) {
        var key = _a[_i];
        var value = config[key];
        if (typeof value === 'boolean') {
            if (value === true) {
                result.push("--" + key);
            }
            continue;
        }
        result.push("--" + key);
        if (Array.isArray(value)) {
            result.push(value.join(','));
        }
        else {
            result.push(value.toString());
        }
    }
    result.push(inFile);
    return result;
}
function parseArray(str) {
    return str.split(',');
}
var args = parser.parseArgs();
if (args.config != null) {
    var configFilePath = args.config;
    if (!ts.sys.fileExists(configFilePath)) {
        throw new Error("Config file \"" + configFilePath + "\" does not exist");
    }
    var configText = ts.sys.readFile(configFilePath);
    if (configText === undefined) {
        throw new Error('Cannot read config file');
    }
    var config = JSON.parse(configText);
    args = parser.parseArgs(configToArgs(config, args.file[0]));
    if (args.outFile != null) {
        args.outFile = path.resolve(configFilePath, '..', args.outFile);
    }
}
if (args.verbose) {
    logger_1.enableVerbose();
}
logger_1.verboseLog("Arguments: " + JSON.stringify(args));
try {
    var inputFilePath = args.file[0];
    var generatedDts = bundle_generator_1.generateDtsBundle(inputFilePath, {
        failOnClass: args.failOnClass,
        outputFilenames: args.outputSourceFileName,
        inlinedLibraries: args.externalInlines ? parseArray(args.externalInlines) : [],
        importedLibraries: args.externalImports ? parseArray(args.externalImports) : [],
        allowedTypesLibraries: args.externalTypes ? parseArray(args.externalTypes) : undefined,
    });
    if (args.outFile == null) {
        var inputFileName = path.parse(inputFilePath).name;
        args.outFile = path.join(inputFilePath, '..', inputFileName + '.d.ts');
    }
    logger_1.normalLog("Writing generated file to " + args.outFile + "...");
    ts.sys.writeFile(args.outFile, generatedDts);
    if (args.noCheck) {
        logger_1.normalLog('Checking of the file is skipped due "no-check" flag');
        process.exit(0);
    }
    logger_1.normalLog('Checking of the generated file...');
    var program = ts.createProgram([args.outFile], get_compiler_options_1.getCompilerOptionsForFile(inputFilePath));
    check_diagnostics_errors_1.checkProgramDiagnosticsErrors(program);
    logger_1.normalLog('Done.');
}
catch (ex) {
    console.error("Error: " + ex.message);
    process.exit(1);
}
